/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014 (c)
 * 
 *    file: numbers.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern int input_format, output_format;
#ifndef __FORMATS
#define __FORMATS
#define DECIMAL_FORMAT	1
#define HEX_FORMAT	2
#define BINARY_FORMAT	3
#define OCTAL_FORMAT	4
#define ALL_FORMATS	5
#endif /* __FORMATS */

char hex_chars[] =
{
  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
  'A', 'B', 'C', 'D', 'E', 'F'
};

long hex_to_decimal(char *hex)
{
  if(strstr(hex, "0X") != 0 && strstr(hex, "0x") != 0)
  {
    printf("Error: Invalid hex format\n");
    return -1;
  }
  long res = 0;
  int i, j, k = (strlen(hex)-3)*4;
  for(j = 2; j < strlen(hex); j++)
  {
    for(i = 0; i < 16; i++)
      if(hex[j] == hex_chars[i] || hex[j] == hex_chars[i]+32)
      { res += (i*(1<<k)); k -= 4; break; }
  }
  return res;
}


long binary_to_decimal(char *binary)
{
  long res = 0;
  int i, j = strlen(binary)-1;
  for(i = 0; i < strlen(binary); i++)
  { res += (((binary[i] == '1') ? 1 : 0) << j); j--; }
  return res;
}


long octal_to_decimal(char *octal)
{
  long res = 0;
  int i, j, k = (strlen(octal)-1)*3;
  for(j = 0; j < strlen(octal); j++)
  {
    i = octal[j]-48;
    res += (i*(1<<k));
    k -= 3;
  }
  return res;
}


/*long decimal_to_octal(long decimal)
{
  long res = 0, remainder = 0, quotient = 8;
  int j = 1;
  while(quotient >= 8)
  {
    quotient = decimal/8;
    remainder = decimal%8;
    res += (remainder*(10*j));
    decimal = quotient;
    j++;
  }
  res += (quotient*(10*j));
  return res;
}*/


void decimal_to_binary(long decimal, char **res)
{
  char res2[64];
  res2[63] = '\0';
  int j = 62, i;
  long quotient = 1, remainder = 0;
  while(quotient)
  {
    quotient = decimal/2;
    remainder = decimal%2;
    res2[j] = remainder ? '1' : '0';
    //printf("q=%ld, r=%ld, j=%d, res2=%c\n", quotient, remainder, j, res2[j]);
    j--;
    decimal = quotient;
  }
  //printf("%d\n", 62-j);
  *res = (char *)malloc(63-j);
  if(!*res)
  {
    printf("Error: Insufficient memory\n");
    return;
  }
  //char res2[62-j];
  int last = 62-j;
  for(i = 0; i < last; i++)
  {
    (*res)[i] = res2[++j];
    //printf("i=%d, res2[%d]=%c\n", i, j, res2[j]);
  }
  (*res)[i] = '\0';
  //return res2;
}


void decimal_to_hex(long decimal, char **res)
{
  char res2[64];
  sprintf(res2, "%x", (unsigned)decimal);
  int pos = 2, i = 0, j = strlen(res2);
  if(j % 2) i = 1;
  *res = (char *)malloc(j+i+2);
  (*res)[0] = '0';
  (*res)[1] = 'x';
  if(i) { (*res)[2] = '0'; pos++; }
  j = 0;
  for(i = 0; i < strlen(res2); i++)
    (*res)[pos++] = res2[j++];
  (*res)[pos] = '\0';
}
/*  char res2[64];
  res2[63] = '\0';
  long remainder = 0, quotient = 16;
  int j = 62, i;
  while(quotient >= 16)
  {
    quotient = decimal/16;
    remainder = decimal%16;
    res2[j] = hex_chars[remainder];
    decimal = quotient;
    j--;
  }
  res2[j] = hex_chars[quotient];
  *res = (char *)malloc(62-j);
  if(!*res)
  {
    printf("Error: Insufficient memory\n");
    return;
  }
  //char res2[62-j];
  (*res)[0] = '0';
  (*res)[1] = 'x';
  for(i = 0; i < 62-j; i++)
    (*res)[i+2] = res2[++j];
  (*res)[i+2] = '\0';
  //return res2;
}*/


void convert_numbers(char *input)
{
  if(input == NULL) return;
  long res;
  if(input[0] == 'b' || input[0] == 'B')
  {
    res = binary_to_decimal(input+1);
    if(output_format == DECIMAL_FORMAT)
      printf("%ld\n", res);
    else if(output_format == OCTAL_FORMAT)
      printf("0%o\n", (unsigned)res);
    else if(output_format == HEX_FORMAT)
    {
      char *res2;
      decimal_to_hex(res, &res2);
      if(res2) { printf("%s\n", res2); free(res2); }
    }
      //printf("%x\n", (unsigned)res);
    else if(output_format == BINARY_FORMAT)
      printf("%s\n", input);
    else if(output_format == ALL_FORMATS)
    {
      printf("%ld\n", res);
      printf("0%o\n", (unsigned)res);
      char *res2;
      decimal_to_hex(res, &res2);
      if(res2) { printf("%s\n", res2); free(res2); }
      //printf("%x\n", (unsigned)res);
      printf("%s\n", input);
    }
  }
  else if((input[0] == '0' && input[1] == 'X') ||
	  (input[0] == '0' && input[1] == 'x'))
  {
    res = hex_to_decimal(input);
    if(output_format == DECIMAL_FORMAT)
      printf("%ld\n", res);
    else if(output_format == OCTAL_FORMAT)
      printf("0%o\n", (unsigned)res);
    else if(output_format == HEX_FORMAT)
      printf("%s\n", input);
    else if(output_format == BINARY_FORMAT)
    {
      char *res2;
      decimal_to_binary(res, &res2);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%ld\n", res);
      printf("0%o\n", (unsigned)res);
      printf("%s\n", input);
      char *res2;
      decimal_to_binary(res, &res2);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
  }
  else if(input[0] == '0')
  {
    res = octal_to_decimal(input);
    if(output_format == DECIMAL_FORMAT)
      printf("%ld\n", res);
    else if(output_format == OCTAL_FORMAT)
      printf("%s\n", input);
    else if(output_format == HEX_FORMAT)
    {
      char *res2;
      decimal_to_hex(res, &res2);
      if(res2) { printf("%s\n", res2); free(res2); }
    }
      //printf("%x\n", (unsigned)res);
    else if(output_format == BINARY_FORMAT)
    {
      char *res2;
      decimal_to_binary(res, &res2);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%ld\n", res);
      printf("%s\n", input);
      char *res2;
      decimal_to_hex(res, &res2);
      if(res2) { printf("%s\n", res2); free(res2); }
      //printf("%x\n", (unsigned)res);
      decimal_to_binary(res, &res2);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
  }
  else
  {
    res = atol(input);
    if(output_format == DECIMAL_FORMAT)
      printf("%ld\n", res);
    else if(output_format == OCTAL_FORMAT)
      printf("0%o\n", (unsigned)res);
    else if(output_format == HEX_FORMAT)
    {
      char *res2;
      decimal_to_hex(res, &res2);
      if(res2) { printf("%s\n", res2); free(res2); }
    }
      //printf("0x%02x\n", (unsigned)res);
    else if(output_format == BINARY_FORMAT)
    {
      char *res2;
      decimal_to_binary(res, &res2);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%ld\n", res);
      printf("0%o\n", (unsigned)res);
      char *res2;
      decimal_to_hex(res, &res2);
      if(res2) { printf("%s\n", res2); free(res2); }
      //printf("%x\n", (unsigned)res);
      decimal_to_binary(res, &res2);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
  }
}